package job_shop_chap_3;

import java.io.File;
import java.io.FileReader;
import java.io.StreamTokenizer;
import java.util.Date;
import org.chocosolver.solver.Model;
import org.chocosolver.solver.Solution;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.search.strategy.Search;
import static org.chocosolver.solver.search.strategy.assignments.DecisionOperatorFactory.makeIntSplit;
import org.chocosolver.solver.search.strategy.selectors.values.IntDomainMiddle;
import org.chocosolver.solver.search.strategy.selectors.values.IntDomainMin;
import org.chocosolver.solver.search.strategy.selectors.variables.FirstFail;
import org.chocosolver.solver.search.strategy.strategy.IntStrategy;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.solver.variables.BoolVar;
import org.chocosolver.util.ESat;



/**
 * 
 * Job-Shop pour le chapitre 3
 */


public class Job_shop_chap3 {
    public static void main(String[] args) {
        int nb_job=-1;			// jobs
        int nb_machine=-1;			// machines
        int Dmax = 999;        	// taille de lâ€™horizon
        int[][] Machine = null;
        int[][] P = null;
        int[][] Tab = null;

        // lecture du fichier
        try{
            File source;
            source = new File(".\\instance\\la10.txt"); // instance lue

            FileReader Input=new FileReader(source);
            StreamTokenizer   Lecteur;
            Lecteur=new StreamTokenizer (Input);
            Lecteur.parseNumbers();

            int nombre_ligne=-1;

            // nb lignes        
            Lecteur.nextToken();
            if (Lecteur.ttype==Lecteur.TT_NUMBER){
                nombre_ligne=(int) Lecteur.nval;
                nb_job=nombre_ligne;
            }

            // nb colonnes
            int nombre_colonne=-1;
            Lecteur.nextToken();
            if (Lecteur.ttype==Lecteur.TT_NUMBER){
                nombre_colonne=(int) Lecteur.nval;
                nb_machine=nombre_colonne;
            }

            // crÃ©ation
            Machine = new int[nb_job][nb_machine];
            P = new int[nb_job][nb_machine];   
            Tab = new int[nb_job][nb_machine];   

            int compteur = 0;

            for (int i=0; i<nombre_ligne; i++){
                for (int j=0; j<nombre_colonne; j++){
                    Tab[i][j] = compteur;
                    compteur++;
                    // lecture machine
                    Lecteur.nextToken();
                    if (Lecteur.ttype==Lecteur.TT_NUMBER){
                        Machine[i][j]=1+(int) Lecteur.nval;
                    }
                    // lecture processing
                    Lecteur.nextToken();
                    if (Lecteur.ttype==Lecteur.TT_NUMBER){
                        P[i][j]=(int) Lecteur.nval;
                    }
                }
            }
            }catch (Exception e) {
                    System.out.println(e.getMessage());
            }; 
            // fin lecture

            Model mon_modele = new Model("JobShop");

            // Contrainte (1.1)
            IntVar[] date_debut = new IntVar[nb_job*nb_machine]; 
            for (int i=0; i<nb_job; i++){
                for (int j=0; j<nb_machine; j++){
                    int position = Tab[i][j];
                    date_debut[position] = mon_modele.intVar("St_"+i+"_"+j, 0,Dmax);
                }
            }
               
            IntVar[] date_fin = new IntVar[nb_job*nb_machine]; 
            for (int i=0; i<nb_job; i++){
                for (int j=0; j<nb_machine; j++){
                    int position = Tab[i][j];
                    date_fin[position] = mon_modele.intVar("Ft_"+i+"_"+j, 0,Dmax);
                }
            }
               
            IntVar[] durees = new IntVar[nb_job*nb_machine];
            for (int i=0; i<nb_job; i++){
                for (int j=0; j<nb_machine; j++){
                    int position = Tab[i][j];
                    durees[position] = mon_modele.intVar("p_"+i+"_"+j, P[i][j]);
                }
            }

            BoolVar[][] b = mon_modele.boolVarMatrix("b",nb_job*nb_machine,nb_job*nb_machine);
            IntVar Cmax = mon_modele.intVar("Cmax", 0, Dmax);

            // contrainte 1. relation entre date de dÃ©but et de fin
            for (int i=0; i<nb_job; i++){
                for (int j=0; j<nb_machine; j++){
                    int cour = Tab[i][j];	
                    mon_modele.arithm(date_debut[cour],"+", durees[cour],"=",date_fin[cour]).post();
                }
            }

            // contrainte 2 : gamme
            for (int i=0; i<nb_job; i++){
                for (int j=1; j<nb_machine; j++){
                    int prec = Tab[i][j-1];
                    int cour = Tab[i][j];
                    mon_modele.arithm(date_fin[prec], "<=",date_debut[cour]).post();
                }
            }

            // contrainte 3 : disjonction 
            for (int i=0; i<nb_job-1; i++){  
                for (int j=0; j<nb_machine; j++){
                    int ope_i = Tab[i][j];
                    for (int u=i+1; u<nb_job; u++){
                        for (int v=0; v<nb_machine; v++){
                            int ope_u = Tab[u][v];
                            if (Machine[i][j]==Machine[u][v]){
                                mon_modele.ifThenElse(b[ope_i][ope_u],
                                    mon_modele.arithm(date_fin[ope_i],"<=",date_debut[ope_u]),
                                    mon_modele.arithm(date_fin[ope_u],"<=",date_debut[ope_i]));
                            }
                        }
                    }
                }
            }

            // fonction objectif
            for (int i = 0; i <nb_job; i++) {
                int valeur = i*nb_machine+nb_machine-1;
                mon_modele.arithm(date_fin[valeur], "<=",Cmax).post();
            }
            
            // Configuration du solveur
            mon_modele.setObjective(Model.MINIMIZE, Cmax);
            Solver mon_solveur = mon_modele.getSolver();
            
            // stratégie par dÃ©faut : Rien faire, branchement sur b
            
            // stratégie du First Fail sur IntDomainMin
             //FirstFail Choix_Variables = new FirstFail(mon_modele);
             //IntDomainMin Choix_Domaine = new IntDomainMin();
             //IntStrategy ma_strategy = Search.intVarSearch(Choix_Variables,Choix_Domaine, date_debut);
            // mon_solveur.setSearch(ma_strategy);

            // Stratégie LDS
               mon_solveur.setLDS(nb_job*nb_machine);
            
            // une autre stratégie
            // mon_solveur.setSearch( Search.intVarSearch (new FirstFail(mon_modele), 
            //             new IntDomainMin(),date_debut));

            // une autre stratégie
            // mon_solveur.setSearch( Search.intVarSearch (new FirstFail(mon_modele), 
            //             new IntDomainMiddle(true),date_debut));
            
            // recherche dichotomique
            // mon_solveur.setSearch(Search.intVarSearch(new FirstFail(mon_modele), 
            //             new IntDomainMiddle(true), makeIntSplit(), date_debut));

            
            Solution s = new Solution(mon_modele);
            Date heure_debut= new Date();
            long h_debut = heure_debut.getTime();

            // affichage des statistiques
            // mon_solveur.showStatistics();
           
            // affichage des short statistiques
            //mon_solveur.showShortStatistics();
            
            // Affichage personnalisÃ©es
            //Mon_message message = new Mon_message( date_debut, b, nb_job, nb_machine, Machine, Tab, date_fin); 
            //mon_solveur.showDecisions(message);
            
            // rÃ©solution 
            while (mon_solveur.solve()) {
                    s.record();
                    int cout=(int) mon_solveur.getBestSolutionValue();
                    Date heure_fin = new Date();
                    long h_fin= heure_fin.getTime();
                    long duree = h_fin-h_debut;
                    long duree_s = duree/1000;
                    System.out.println("temps : " + duree_s + " s" + "    Cmax = " + cout);
            }  

            // affichage de la solution
            if (mon_solveur.isFeasible() == ESat.TRUE){
                    int cout=(int) mon_solveur.getBestSolutionValue();
                    System.out.println("Cmax = " + cout);
                    for (int i=0; i<nb_job; i++){
                            for(int j=0; j<nb_machine; j++){
                                    int position = Tab[i][j];
                                    int deb = s.getIntVal(date_debut[position]);
                                    int fin = s.getIntVal(date_fin[position]);
                                    System.out.print(deb + "/" + fin);
                                    System.out.print("    ");
                            }
                            System.out.println();
                    }
            }
            else
                    System.out.println("pas de sol.faisable");
        }
}
    
    

